/*
Copyright (c) 2008, Jari Komppa

All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, 
  this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation 
  and/or other materials provided with the distribution.

* Neither the name of the library's copyright owner nor the names of its 
  contributors may be used to endorse or promote products derived from this 
  software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include "nelumbo.h"

FLOAT linear(FLOAT t, FLOAT p0, FLOAT p1, FLOAT p2, FLOAT p3)
{
    return p1 + (p2-p1) * t;
}


FLOAT catmullrom(FLOAT t, FLOAT p0, FLOAT p1, FLOAT p2, FLOAT p3)
{    
    return FLOAT(0.5f) * (
           (2 * p1) + 
           (-p0 + p2) * t + 
           (2 * p0 - 5 * p1 + 4 * p2 - p3) * t * t + 
           (-p0 + 3 * p1 - 3 * p2 + p3) * t * t * t 
           );
}

FLOAT bspline(FLOAT t, FLOAT p0, FLOAT p1, FLOAT p2, FLOAT p3)
{
    FLOAT it = FLOAT(1) - t;

		FLOAT b0 = it*it*it/FLOAT(6.0f);
		FLOAT b1 = (3*t*t*t - 6*t*t +4)/FLOAT(6.0f);
		FLOAT b2 = (-3*t*t*t +3*t*t + 3*t + 1)/FLOAT(6.0f);
		FLOAT b3 =  t*t*t/FLOAT(6.0f);
    return     
        b0*p0 +
        b1*p1 +
        b2*p2 +
        b3*p3;
}


void drawspan(int y, int x1, int w, int c1, int c2)
{
    int i;
    if (x1 > 319)
        x1 = 319;
    if (x1 < 0)
    {
        w += x1;
        x1 = 0;
    }
    if (x1+w > 319)
    {
        w = 319 - x1;
    }
    int yofs = y * gScreen->pitch / 2;    
    for (i = 0; i < x1; i++, yofs++)
        *(((unsigned short*)gScreen->pixels) + yofs) = c2;
    for (; i < x1+w; i++, yofs++)
        *(((unsigned short*)gScreen->pixels) + yofs) = c1;
    for (; i < 320; i++, yofs++)
        *(((unsigned short*)gScreen->pixels) + yofs) = c2;
/*
    for (i = 0; i < 320; i++)
    {
        if (i < x1 || i > x1+w)
            *(((unsigned short*)gScreen->pixels) + yofs + i) = c2;
        else
            *(((unsigned short*)gScreen->pixels) + yofs + i) = c1;
    }
    */
}

void persp(FLOAT &x, FLOAT &y, FLOAT &w, FLOAT z,FLOAT fov)
{
    x = (x*fov / (z + 1)) + 160;
    w = (w*fov / (z + 1));
    y = (y*fov / (z + 1)) + 120;
}

void trackpick(FLOAT &x, FLOAT &y, FLOAT &z, FLOAT d)
{
    // note: assumes that all segments have the same length - non-constant velocity
    d /= (FLOAT(currentlevel->tracklen) / FLOAT(currentlevel->controlpoints));
    FLOAT segpos = d - FLOOR(d);
    int p1 = TOINT(d) % currentlevel->controlpoints;
    int p0 = (p1 + currentlevel->controlpoints - 1) % currentlevel->controlpoints;
    int p2 = (p1 + 1) % currentlevel->controlpoints;
    int p3 = (p1 + 2) % currentlevel->controlpoints;
    p0 *= 3;
    p1 *= 3;
    p2 *= 3;
    p3 *= 3;

    FLOAT *track = currentlevel->trackdata;
/*
    x = linear(segpos,track[p0],track[p1],track[p2],track[p3]);
    track++;
    y = linear(segpos,track[p0],track[p1],track[p2],track[p3]);
    track++;
    z = linear(segpos,track[p0],track[p1],track[p2],track[p3]);
*/
    x = catmullrom(segpos,track[p0],track[p1],track[p2],track[p3]);
    track++;
    y = catmullrom(segpos,track[p0],track[p1],track[p2],track[p3]);
    track++;
    z = catmullrom(segpos,track[p0],track[p1],track[p2],track[p3]);

}

void rotate_around_y(FLOAT &x, FLOAT &z, FLOAT sa, FLOAT ca)
{
    // 2d rotation:
    //z' = z cos f - x sin f 
    //x' = z sin f + x cos f   
    
    FLOAT x1, z1;
    
    x1 = x;
    z1 = z;

    z = z1 * ca - x1 * sa;
    x = z1 * sa + x1 * ca;
}

void rotate_around_x(FLOAT &y, FLOAT &z, FLOAT sa, FLOAT ca)
{
    // 2d rotation:
    //z' = y cos f - z sin f 
    //y' = y sin f + z cos f   
    
    FLOAT y1, z1;
    
    y1 = y;
    z1 = z;
 
    y = ca*y1 - sa*z1;
    z = sa*y1 + ca*z1;
}

void drawrect(int x, int y, int w, int h, int color)
{
	SDL_Rect r;
	r.x = x;
	r.y = y;
	r.w = w;
	r.h = h;
	SDL_FillRect(gScreen, &r, color);
}

void blitbg(FLOAT dir, int tilt, int sl)
{
    drawrect(0,0,320,sl,0);
	if (tilt < 0) sl += tilt;
    if (sl <= 0) return;
    int xofs = TOINT(((dir) * FLOAT(0.5f/M_PI)) * gBg->w);
	SDL_Rect src, dst;
	src.w = gBg->w;
    src.h = sl;
	src.x = 0;
	src.y = tilt;
	dst.w = 320;
	dst.h = sl;
	dst.x = xofs;
	dst.y = 0;
	SDL_BlitSurface(gBg, &src, gScreen, &dst);        
    dst.x -= gBg->w;
	SDL_BlitSurface(gBg, &src, gScreen, &dst);
}

struct delayedsprite
{
    FLOAT x, y;
    SDL_Surface *pic;
    int w, h;
    int color;
};

delayedsprite sprite[10000];
int maxsprite;

/*
fixed_point<16> atan2(fixed_point<16> x, fixed_point<16> y)
{
    fixed_point<16> p = x / y;
    fixed_point<16> p3 = p*p*p;
    fixed_point<16> p5 = p3*p*p;
    fixed_point<16> p7 = p5*p*p;
    fixed_point<16> p9 = p7*p*p;

    return 1*p
        -fixed_point<16>(1/3.0f)*p3
        +fixed_point<16>(1/5.0f)*p5
        -fixed_point<16>(1/7.0f)*p7
        +fixed_point<16>(1/9.0f)*p9;
}
*/

void drawsprite(SDL_Surface *surf, FLOAT x, FLOAT y, int w, int h, int c)
{
    int ix = TOINT(x);
    int iy = TOINT(y);
    if (w == 0 || ix + w < 0 || iy + h < 0 || ix > 319 || iy > 239) return;

    int step = ((surf->w-1) * (1<<16)) / w;
    int maxh = ((surf->h-1) * w) / (surf->w-1);

    int xofs = 0;
    int yofs = 0;

#ifdef USE_FLOATS
    int xfract = TOINT((x - FLOOR(x))*(1 << 16));
    int yfract = TOINT((y - FLOOR(y))*(1 << 16));
#else
    int xfract = (x - FLOOR(x)).intValue;
    int yfract = (y - FLOOR(y)).intValue;
#endif

    if (ix < 0)
    {
        xofs = -ix;
        w += ix;
        ix = 0;
    }

    if (iy < 0)
    {
        yofs = -iy;
        h += iy;
        iy = 0;
    }

    if (ix + w > 319)
    {
        w = 319 - ix;
    }

    if (iy + h > 239)
    {
        h = 239 - iy;
    }

    int dstpitch = gScreen->pitch / 2;
    unsigned short * dstptrofs = (unsigned short *)gScreen->pixels + ix + iy * dstpitch;
    unsigned short * dstptr = dstptrofs;

    int srcpitch = surf->pitch / 2;
    unsigned short * srcptr = (unsigned short *)surf->pixels;

    int xp = 0;
    int yp = yofs * step + yfract;

    if (yofs + h >= maxh) 
        h = maxh - yofs;

    int i, j;
    for (i = 0; i < h; i++)
    {
        dstptr = dstptrofs;
        srcptr = (unsigned short *)surf->pixels + (yp >> 16) * srcpitch;
        xp = xofs * step + xfract;
        for (j = 0; j < w; j++)
        {
            int col = srcptr[xp >> 16];
            if (col != RGB(255,255,0))
            {
                /*
                int b = ((col >> 0) & 0x1f) << 3;
                int g = ((col >> 6) & 0x1f) << 3;
                int r = ((col >> 11) & 0x1f) << 3;
                *dstptr = RGB(COLORMIX(r,0,c),
                              COLORMIX(g,0,c),
                              COLORMIX(b,0,c));
                              */
                *dstptr = col;
            }
            xp += step;
            dstptr++;
        }
        dstptrofs += dstpitch;
        yp += step;
    }

    //    drawrect(x,y,w,h,c);

}


FLOAT pos = 0;

void render_ingame()
{   
  // Lock surface if needed
  if (SDL_MUSTLOCK(gScreen))
    if (SDL_LockSurface(gScreen) < 0) 
      return;

  int i;

  // Ask SDL for the time in milliseconds
  unsigned int tick = SDL_GetTicks(); 

  while (gLastTick < tick)
  {
      pos += 0.2f;
      gLastTick += 10;
  }

  FLOAT step = 0.1f;
//  FLOAT pos = tick / 100.0f;
/*
  char temp[200];
  sprintf(temp,"%3.3f",tick /100.0f);
  SDL_WM_SetCaption(temp, NULL);  
*/
  FLOAT x,y,z, x1,y1,z1;
  trackpick(x,y,z,pos + step*2);
  y1 = y;
  trackpick(x1,y,z1,pos + 10);//step); // lookat
  trackpick(x,y,z,pos+step);
  z1 = z1 - z;
  y1 = y1 - y;
  x1 = x1 - x;
  FLOAT yrot = //-atan2(x1,z1);
      -atan2(TOFLOAT(x1), TOFLOAT(z1)); // TODO
  FLOAT xrot = y1 * 10 + FLOAT(0.5f);
  int lastscanline = 239;
  FLOAT rz = 1;
  maxsprite = 0;
  FLOAT nextsprite = FLOOR(pos/10)*10;

#ifdef CONSOLE_DEBUG
  int START;
#endif

  TIMING_START;

  FLOAT cax = cos(xrot);
  FLOAT cay = cos(yrot);
  FLOAT sax = sin(xrot);
  FLOAT say = sin(yrot);
  FLOAT fov = 320.0f / tanf((60.0f / 180.0f) * float(M_PI));
  FLOAT xpos = 0;//sin(tick / 1000.0f) * 2;
/*
  vec3<FLOAT> eye;
  vec3<FLOAT> at;
  vec3<FLOAT> up;
  eye.x = 0;
  eye.y = 0+1;
  eye.z = 0;
  at.x = -x1;
  at.y = -y1;
  at.z = -z1;
  up.x = 0;
  up.y = 1;
  up.z = 0;

  mat4<FLOAT> lookat = lookat_matrix<FLOAT>(eye,at,up);
*/

  for (i = 1; i < 500; i++)
  {
      FLOAT xp, yp, zp, w;
      trackpick(xp, yp, zp, pos + rz);
      rz += i * (1 / 2000.0f);
/*
      vec4<FLOAT> v;
      v.x = xp;
      v.y = yp;
      v.z = zp;
      v.w = 1;
      v = lookat * v;
*/
      xp -= x;
      yp -= y;
      zp -= z;
      rotate_around_y(xp,zp,say,cay);
      rotate_around_x(yp,zp,sax,cax);
      xp += xpos;
      yp += 2;
      w = 1;
      
//      xp=v.x;yp=v.y;zp=v.z;

      if (zp > -1)
      {
        persp(xp,yp,w,zp,fov);
        if (pos + rz > nextsprite)
        {          
            nextsprite+=1;
            if (pos + rz < nextsprite)
            {

                if (TOINT(nextsprite) & 1)
                    sprite[maxsprite].x = (xp + 2 * w);
                else
                    sprite[maxsprite].x = (xp - 4 * w);
                sprite[maxsprite].y = (yp - 2 * w);
                sprite[maxsprite].w = TOINT(w * 2);
                sprite[maxsprite].h = lastscanline - TOINT(sprite[maxsprite].y);
                sprite[maxsprite].color = i/2;
                sprite[maxsprite].pic = gTreetest;                
                if (sprite[maxsprite].h > 0)
                    maxsprite++;

                gRand.init_genrand(TOINT(nextsprite));
                sprite[maxsprite].x = xp + FLOAT(gRand.genrand_real1() * 38.0f - 20.0f) * w ;
                sprite[maxsprite].y = (yp - 4 * w);
                sprite[maxsprite].w = TOINT(w * 2);
                sprite[maxsprite].h = lastscanline - TOINT(sprite[maxsprite].y);
                sprite[maxsprite].color = i/2;
                sprite[maxsprite].pic = gCloudtest;                
                if (sprite[maxsprite].h > 0)
                    maxsprite++;

                if ((TOINT(nextsprite) & 15) == 0)
                {
                    sprite[maxsprite].x = xp - 1 * w;
                    sprite[maxsprite].y = (yp - FLOAT(0.5f) * w);
                    sprite[maxsprite].w = TOINT(w * 1);
                    sprite[maxsprite].h = lastscanline - TOINT(sprite[maxsprite].y);
                    sprite[maxsprite].color = i/2;
                    sprite[maxsprite].pic = gCartest[1];                
                    if (sprite[maxsprite].h > 0)
                        maxsprite++;
                }


                if ((TOINT(nextsprite) % currentlevel->tracklen) == 0)
                {
                    sprite[maxsprite].x = (xp - 2 * w);
                    sprite[maxsprite].y = (yp - FLOAT(2.5f) * w);
                    sprite[maxsprite].w = TOINT(w * 4);
                    sprite[maxsprite].h = lastscanline - TOINT(sprite[maxsprite].y);
                    sprite[maxsprite].color = i/2;
                    sprite[maxsprite].pic = gGoaltest;                
                    if (sprite[maxsprite].h > 0)
                        maxsprite++;
                }

            }
        }
        w *= 4;
        if (yp < 0) yp = 0;
        if (lastscanline && yp < lastscanline)
        {
            int dark = TOINT(pos + rz) & 2;
            int cr = (dark)?RGB(COLORMIX(40,0,i/2),COLORMIX(40,0,i/2),COLORMIX(40,0,i/2)):RGB(COLORMIX(60,0,i/2),COLORMIX(60,0,i/2),COLORMIX(60,0,i/2));
            int cs = (dark)?RGB(0,COLORMIX(100,0,i/2),0):RGB(0,COLORMIX(200,0,i/2),0);              
            
            for (; lastscanline >= TOINT(yp); lastscanline--)
            {
                drawspan(
                    lastscanline, 
                    TOINT(xp - w/2),
                    TOINT(w),
                    cr, 
                    cs);
            }
        }
      }
  }

  TIMING_END("ground");

  TIMING_START;
  if (lastscanline > 0)
    blitbg(yrot,TOINT(xrot * 100),lastscanline+1);
  TIMING_END("bg");

  TIMING_START;
  for (i = maxsprite-1; i >= 0; i--)
  {
      drawsprite(sprite[i].pic, sprite[i].x,sprite[i].y,sprite[i].w,sprite[i].h,sprite[i].color);
  }

  drawsprite(gCartest[1], 140,160,80,80,0xffff);

  
  TIMING_END("sprites");


  TIMING_START;
  
  // draw map
	SDL_Rect src, dst;
	src.w = 88;
    src.h = 88;
	src.x = 0;
	src.y = 0;
	dst.w = 88;
	dst.h = 88;
	dst.x = 232;
	dst.y = 0;
	SDL_BlitSurface(currentlevel->map, &src, gScreen, &dst);        
  
    trackpick(x, y, z, pos);
    x -= currentlevel->mapxofs;
    z -= currentlevel->mapzofs;
    x = (x * 80) / currentlevel->mapscale;
    z = (z * 80) / currentlevel->mapscale;
    x = 80 - x;
    drawrect(TOINT(x)+240-6, TOINT(z)+2, 6,6,RGB(255,0,0));
    char maplen[200];
    sprintf(maplen,"%3.2fKM",currentlevel->tracklen*0.01f);
    drawstring(maplen,240,100);

    TIMING_END("map");

  // Unlock if needed
  if (SDL_MUSTLOCK(gScreen)) 
    SDL_UnlockSurface(gScreen);


#define PERF_FRAMES 50

    static int perf_idx = 0;
    static int perf_data[PERF_FRAMES];    
    perf_data[perf_idx] = tick;
#ifdef CONSOLE_DEBUG
    printf("%3.3f mspf (%3.3f fps)\n", (float)(perf_data[perf_idx] - perf_data[(perf_idx+1) % PERF_FRAMES]) / PERF_FRAMES,
                                       1000.0f / ((float)(perf_data[perf_idx] - perf_data[(perf_idx+1) % PERF_FRAMES]) / PERF_FRAMES));
#endif
    char temp[60];
    sprintf(temp,"%3.1f FPS", 1000.0f / ((float)(perf_data[perf_idx] - perf_data[(perf_idx+1) % PERF_FRAMES]) / PERF_FRAMES));
    drawstring(temp,5,5);

    sprintf(temp,"DISTANCE:%3.1f%%", (TOFLOAT(pos) / (currentlevel->tracklen))*100.0f);
    drawstring(temp,5,220);

    perf_idx++;
    perf_idx %= PERF_FRAMES;

   // Tell SDL to update the whole gScreen
  SDL_UpdateRect(gScreen, 0, 0, 320, 240);    
}

void key_ingame(int key, int down)
{
    if (down==0)
        gGameState = 0;
}
