/*
Copyright (c) 2008, Jari Komppa

All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, 
  this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation 
  and/or other materials provided with the distribution.

* Neither the name of the library's copyright owner nor the names of its 
  contributors may be used to endorse or promote products derived from this 
  software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include "nelumbo.h"

leveldata::leveldata()
{
    trackdata = 0;
    staticsprite = 0;
    mobilesprite = 0;
    trackseglen = 0;
}

leveldata::~leveldata()
{
    delete[] trackdata;
    delete[] staticsprite;
    delete[] mobilesprite;
    delete[] trackseglen;
    SDL_FreeSurface(map);
};

void drawmaprect(int x, int y, int w, int h, int color)
{
	SDL_Rect r;
	r.x = x;
	r.y = y;
	r.w = w;
	r.h = h;
	SDL_FillRect(currentlevel->map, &r, color);
}

void trackpicklin(FLOAT &x, FLOAT &y, FLOAT &z, FLOAT d)
{
    int seg = 0;
    while (d > currentlevel->trackseglen[seg])
    {
        d -= currentlevel->trackseglen[seg];
        seg++;
    }
    FLOAT segpos = (d / currentlevel->trackseglen[seg]);

    int p1 = seg % currentlevel->controlpoints;
    int p0 = (p1 + currentlevel->controlpoints - 1) % currentlevel->controlpoints;
    int p2 = (p1 + 1) % currentlevel->controlpoints;
    int p3 = (p1 + 2) % currentlevel->controlpoints;
    p0 *= 3;
    p1 *= 3;
    p2 *= 3;
    p3 *= 3;

    FLOAT *track = currentlevel->trackdata;

    x = bspline(segpos,track[p0],track[p1],track[p2],track[p3]);
    track++;
    y = bspline(segpos,track[p0],track[p1],track[p2],track[p3]);
    track++;
    z = bspline(segpos,track[p0],track[p1],track[p2],track[p3]);
}


FLOAT bsplineseglen(FLOAT xp0, FLOAT xp1, FLOAT xp2, FLOAT xp3, FLOAT yp0, FLOAT yp1, FLOAT yp2, FLOAT yp3)
{
    FLOAT totallen = 0;
    int i;
    FLOAT x = bspline(0,xp0,xp1,xp2,xp3);
    FLOAT y = bspline(0,yp0,yp1,yp2,yp3);
    for (i = 1; i < 20; i++)
    {
        FLOAT x1 = bspline(i * FLOAT(1.0f / 100),xp0,xp1,xp2,xp3);
        FLOAT y1 = bspline(i * FLOAT(1.0f / 100),yp0,yp1,yp2,yp3);
        FLOAT xd = x1-x;
        FLOAT yd = y1-y;
        FLOAT len = sqrt(xd*xd+yd*yd);
        totallen += len * 5;
        x = x1;
        y = y1;
    }

    return totallen;
}

// not used
void smooth()
{
    leveldata * cl = currentlevel;

    FLOAT *seglen = cl->trackseglen;
    int i,j;

    for (j = 0; j < 10; j++)
    {
        FLOAT lensum = 0;
        for (i = 0; i < cl->controlpoints; i++)
        {
//            FLOAT xd = cl->trackdata[i*3+0] - cl->trackdata[((i + 1)%cl->controlpoints)*3+0];
//            FLOAT zd = cl->trackdata[i*3+2] - cl->trackdata[((i + 1)%cl->controlpoints)*3+2];
//            seglen[i] = sqrt(xd*xd+zd*zd);
            seglen[i] = bsplineseglen(
                cl->trackdata[((i + cl->controlpoints - 1) % cl->controlpoints)*3+0],
                cl->trackdata[((i) % cl->controlpoints)*3+0],
                cl->trackdata[((i + 1) % cl->controlpoints)*3+0],
                cl->trackdata[((i + 2) % cl->controlpoints)*3+0],
                cl->trackdata[((i + cl->controlpoints - 1) % cl->controlpoints)*3+2],
                cl->trackdata[((i) % cl->controlpoints)*3+2],
                cl->trackdata[((i + 1) % cl->controlpoints)*3+2],
                cl->trackdata[((i + 2) % cl->controlpoints)*3+2]);
            lensum += seglen[i];
        }
        for (i = 0; i < cl->controlpoints; i++)
            seglen[i] /= lensum;
        for (i = 0; i < cl->controlpoints; i++)
        {
            FLOAT x,y,z;
            trackpicklin(x,y,z,i*(FLOAT(1.0f) / FLOAT(cl->controlpoints)));
            cl->trackdata[i * 3 + 0] = x;
            cl->trackdata[i * 3 + 2] = z;
        }        

        cl->tracklen = TOINT(lensum);
    }
}

void remap()
{
    leveldata * cl = currentlevel;
    int i;
    int newcpcount = cl->controlpoints * 20;
    FLOAT *newcp = new FLOAT[newcpcount * 3];   
    delete[] cl->trackseglen;
    cl->trackseglen = new FLOAT[cl->controlpoints];
    FLOAT *seglen = cl->trackseglen;


    FLOAT lensum = 0;
    for (i = 0; i < cl->controlpoints; i++)
    {
        seglen[i] = bsplineseglen(
            cl->trackdata[((i + cl->controlpoints - 1) % cl->controlpoints)*3+0],
            cl->trackdata[((i) % cl->controlpoints)*3+0],
            cl->trackdata[((i + 1) % cl->controlpoints)*3+0],
            cl->trackdata[((i + 2) % cl->controlpoints)*3+0],
            cl->trackdata[((i + cl->controlpoints - 1) % cl->controlpoints)*3+2],
            cl->trackdata[((i) % cl->controlpoints)*3+2],
            cl->trackdata[((i + 1) % cl->controlpoints)*3+2],
            cl->trackdata[((i + 2) % cl->controlpoints)*3+2]);
        lensum += seglen[i];
    }
    for (i = 0; i < cl->controlpoints; i++)
        seglen[i] /= lensum;
    cl->tracklen = 1;
    for (i = 0; i < newcpcount; i++)
    {
        FLOAT x,y,z;
        trackpicklin(x,y,z,i * (FLOAT(1.0f) / FLOAT(newcpcount)));
        newcp[i * 3 + 0] = x;
        newcp[i * 3 + 1] = y;
        newcp[i * 3 + 2] = z;
    }            
    cl->tracklen = TOINT(lensum);
    delete[] seglen;
    cl->trackseglen = new FLOAT[newcpcount];
    delete[] cl->trackdata;
    cl->trackdata = newcp;
    cl->controlpoints = newcpcount;
}

int notvalidseg(int segno)
{
    leveldata * cl = currentlevel;
    if (segno == 0)
        return 0;
    int i;
    for (i = 0; i < segno; i++)
    {
        if (cl->trackdata[i * 3 + 0] == cl->trackdata[segno * 3 + 0] ||
            cl->trackdata[i * 3 + 2] == cl->trackdata[segno * 3 + 2])
            return 1;
    }
    return 0;
}

/*
great for linear tracks..
    FLOAT lastdir = 0;
    for (i = 1; i < currentlevel->controlpoints; i++)
    {
        FLOAT dir = lastdir + gRand.genrand_real1() * M_PI - M_PI * 0.5; //FLOAT(gRand.genrand_real1() * M_PI * 2);
        lastdir = dir;

        cl->trackdata[i*3+0] = cl->trackdata[(i - 1)*3+0] + sin(dir) * 50;
        cl->trackdata[i*3+1] = 0;
        cl->trackdata[i*3+2] = cl->trackdata[(i - 1)*3+2] + cos(dir) * 50;
    }

*/

int fudge(int i)
{    
    int fudging;
    int fudged = 0;

    do 
    {
        fudging = 0;
        leveldata * cl = currentlevel;
        FLOAT x1 = cl->trackdata[i*3+0] - cl->trackdata[((i+1)%cl->controlpoints)*3+0];
        FLOAT y1 = cl->trackdata[i*3+2] - cl->trackdata[((i+1)%cl->controlpoints)*3+2];
        FLOAT x2 = cl->trackdata[i*3+0] - cl->trackdata[((i+cl->controlpoints-1)%cl->controlpoints)*3+0];
        FLOAT y2 = cl->trackdata[i*3+2] - cl->trackdata[((i+cl->controlpoints-1)%cl->controlpoints)*3+2];
        FLOAT d1 = sqrt(x1*x1+y1*y1);
        FLOAT d2 = sqrt(x2*x2+y2*y2);
        x1 /= d1;
        y1 /= d1;
        x2 /= d2;
        y2 /= d2;

        FLOAT dot = x1*x2+y1*y2;

        if (d1 < 10 || d2 < 10 || dot > 0)
        {
            // maybe we should do something sane instead of this
            cl->trackdata[i*3+0] = FLOAT(gRand.genrand_real1()) * FLOAT(500);
            cl->trackdata[i*3+2] = FLOAT(gRand.genrand_real1()) * FLOAT(500);
            fudged = 1;
    //        fudging = 1;
        }
    }
    while (fudging);

    return fudged;
}

void genlevel(int basetype, int scenario, int seed)
{
    delete currentlevel;
    currentlevel = new leveldata;
    leveldata * cl = currentlevel;

    cl->controlpoints = 10;
    cl->tracklen = cl->controlpoints * 20;
    cl->trackdata = new FLOAT[cl->controlpoints * 3];

    // Generate some base points..

    gRand.init_genrand(seed+6);
    int i, j;

    FLOAT minx,minz,maxx,maxz,miny,maxy;
    FLOAT dx,dy,dz;

    cl->trackdata[0] = cl->trackdata[1] = cl->trackdata[2] = 0;

    for (i = 1; i < currentlevel->controlpoints; i++)
    {
        cl->trackdata[i*3+0] = FLOAT(gRand.genrand_real1()) * FLOAT(500);
            //FLOOR(sin(i * FLOAT((M_PI * 2)/cl->controlpoints)) * 5)*25;
        cl->trackdata[i*3+1] = 0;
        cl->trackdata[i*3+2] = FLOAT(gRand.genrand_real1()) * FLOAT(500);
            //FLOOR(cos(i * FLOAT((M_PI * 2)/cl->controlpoints)) * 5)*25;
    }

    // control points may cause too steep curves, so we'll fudge them until they don't.

    j = 1;
    while (j)
    {
        j = 0;
        int ofs = gRand.genrand_int31() % cl->controlpoints;
        for (i = 0; i < cl->controlpoints; i++)
        {
            j |= fudge((i + ofs) % cl->controlpoints);
        }
    }
   
    // now the spline doesn't have constant speed, so we'll normalize it..
    // (idea based on an article in game programming gems, too lazy to look up the 
    // reference - constant speed splines anyway)

    remap();
    
    remap();


    // hills

    int hillcount = gRand.genrand_int31() % 100;

    if (cl->tracklen > 200)
    {
        for (i = 0; i < hillcount; i++)
        {
            int width = TOINT(gRand.genrand_real1() * cl->controlpoints * FLOAT(0.5f) + (FLOAT(cl->tracklen) / FLOAT(cl->controlpoints)) * 5000);
            int offset = TOINT(gRand.genrand_real1() * cl->controlpoints);
            FLOAT height = FLOAT(cl->tracklen) / 200;//FLOAT(gRand.genrand_real1() * 5);
            for (j = 0; j < width; j++)
            {
                FLOAT s = sin(j * FLOAT(M_PI / width));
                s = s * s * s * s; // nice bell curve
                cl->trackdata[((j + offset)%cl->controlpoints)*3+1] -= s * height;
            }
        }
    }

    /*
    
    for (i = 0; i < currentlevel->controlpoints; i++)
        cl->trackdata[i*3+1] = gRand.genrand_real1() * 20;

    for (j = 0; j < 1000; j++)
    for (i = 0; i < currentlevel->controlpoints; i++)
        cl->trackdata[i*3+1] = ((cl->trackdata[i*3+1] * 2) + 
                                cl->trackdata[((i + currentlevel->controlpoints - 1)%currentlevel->controlpoints)*3+1] +
                                cl->trackdata[((i + currentlevel->controlpoints + 1)%currentlevel->controlpoints)*3+1]) / 4;
*/


    // generate map
    trackpick(minx, miny, minz, 0);
    maxz = minz;
    maxx = minx;
    maxy = miny;

    for (i = 0; i < cl->tracklen; i++)
    {
        FLOAT x,y,z;
        trackpick(x, y, z, FLOAT(i));
        if (minx > x) minx = x;
        if (minz > z) minz = z;
        if (miny > y) miny = y;
        if (maxy < y) maxy = y;
        if (maxx < x) maxx = x;
        if (maxz < z) maxz = z;
    }
    dx = maxx-minx;
    dz = maxz-minz;
    dy = maxy-miny;
    FLOAT scale = dx;
    if (dz > dx) scale = dz;
    cl->mapscale = scale;
    cl->mapxofs = minx;
    cl->mapzofs = minz;

    cl->map = SDL_CreateRGBSurface(SDL_SWSURFACE, 88, 88, 16, RGB(255,0,0), RGB(0,255,0), RGB(0,0,255),0);

    for (i = 0; i < 88; i++)
    {
        for (j = 0; j < 88; j++)
        {
            ((unsigned short*)cl->map->pixels)[j * cl->map->pitch / 2 + i] = RGB(255,255,0);
        }
    }

    FLOAT x=0,y=0,z=0;
    for (i = 0; i < cl->tracklen; i++)
    {
        trackpick(x, y, z, FLOAT(i));
        x -= minx;
        z -= minz;
        x /= scale;
        z /= scale;
        x *= 80;
        z *= 80;
        x = 80 - x;
        int c = 255 - TOINT(((y - miny) / dy) * 128);
        drawmaprect(TOINT(x)+3,TOINT(z)+3,3,3,RGB(c,c,c));//(255,255,255));
    }
    
    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
            drawmaprect(TOINT(x+(i-1)*3),TOINT(z+(j-1)*3),3,3,RGB(255,255,255) * ((i&1)^(j&1)));
   
    SDL_SetColorKey(cl->map, SDL_SRCCOLORKEY, RGB(255,255,0));  
}
