/*
Copyright (c) 2008, Jari Komppa

All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, 
  this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation 
  and/or other materials provided with the distribution.

* Neither the name of the library's copyright owner nor the names of its 
  contributors may be used to endorse or promote products derived from this 
  software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include "nelumbo.h"

MersenneTwister gRand;

SDL_Surface *gScreen;
SDL_Surface *gBg;
SDL_Surface *gTreetest;
SDL_Surface *gCloudtest;
SDL_Surface *gCartest[3];
SDL_Surface *gFont;
SDL_Surface *gMenubg;
SDL_Surface *gGoaltest;

leveldata *currentlevel = NULL;

unsigned int gLastTick = 0;
int gGameState = 0;

void drawchar(char ch, int x, int y)
{
	SDL_Rect src, dst;
	src.w = 10;
	src.h = 16;
	src.x = 0;
	src.y = (ch - 32) * 16;
	dst.w = 10;
	dst.h = 16;
	dst.x = x;
	dst.y = y;
	SDL_BlitSurface(gFont, &src, gScreen, &dst);
}

void drawstring(const char * string, int x, int y)
{
	while (*string)
	{
		drawchar(*string,x,y);
		x += 10;
		string++;
	}
}


void render()
{
    if (gGameState)
        render_ingame();
    else
        render_menu();
}

void handle_key(int key, int down)
{
    if (gGameState)
        key_ingame(key, down);
    else
        key_menu(key, down);
}

void gp2x_cleanup()
{
    delete currentlevel;
    SDL_FreeSurface(gScreen);
    SDL_FreeSurface(gBg);
    SDL_FreeSurface(gTreetest);
    SDL_FreeSurface(gCloudtest);
    SDL_FreeSurface(gCartest[0]);
    SDL_FreeSurface(gCartest[1]);
    SDL_FreeSurface(gCartest[2]);
    SDL_FreeSurface(gFont);
    SDL_FreeSurface(gMenubg);
    SDL_FreeSurface(gGoaltest);
//    SDL_CloseAudio();
#ifndef _MSC_VER
    chdir("/usr/gp2x");
	execl("/usr/gp2x/gp2xmenu", "/usr/gp2x/gp2xmenu", NULL);
#endif
}


// Entry point
int main(int argc, char *argv[])
{
    printf("Here we go..\n");
    int i, j;

  // Initialize SDL's subsystems - in this case, only video.
  if ( SDL_Init(SDL_INIT_JOYSTICK | SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_TIMER) < 0 )
//  if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) 
  {
    fprintf(stderr, "Unable to init SDL: %s\n", SDL_GetError());
    exit(1);
  }

  // Register SDL_Quit to be called at exit; makes sure things are
  // cleaned up when we quit.
#ifdef _MSC_VER    
    atexit(SDL_Quit);
#else
    atexit(gp2x_cleanup);
    SDL_JoystickOpen(0);
#endif
   
  // Attempt to create a 640x480 window with 32bit pixels.
  gScreen = SDL_SetVideoMode(320, 240, 16, SDL_SWSURFACE);

#ifndef _MSC_VER    
    SDL_ShowCursor(0);
#endif

  SDL_Surface *temp = SDL_LoadBMP("testbg.bmp");
  gBg = SDL_ConvertSurface(temp, gScreen->format,SDL_SWSURFACE);
  SDL_FreeSurface(temp);

  temp = SDL_LoadBMP("treetest.bmp");
  gTreetest = SDL_ConvertSurface(temp, gScreen->format,SDL_SWSURFACE);
  SDL_FreeSurface(temp);

  temp = SDL_LoadBMP("testcloud.bmp");
  gCloudtest = SDL_ConvertSurface(temp, gScreen->format,SDL_SWSURFACE);
  SDL_FreeSurface(temp);

  temp = SDL_LoadBMP("testcar0.bmp");
  gCartest[0] = SDL_ConvertSurface(temp, gScreen->format,SDL_SWSURFACE);
  SDL_FreeSurface(temp);

  temp = SDL_LoadBMP("testcar1.bmp");
  gCartest[1] = SDL_ConvertSurface(temp, gScreen->format,SDL_SWSURFACE);
  SDL_FreeSurface(temp);

  temp = SDL_LoadBMP("testcar2.bmp");
  gCartest[2] = SDL_ConvertSurface(temp, gScreen->format,SDL_SWSURFACE);
  SDL_FreeSurface(temp);

  temp = SDL_LoadBMP("goaltest.bmp");
  gGoaltest = SDL_ConvertSurface(temp, gScreen->format,SDL_SWSURFACE);
  SDL_FreeSurface(temp);

  temp = SDL_LoadBMP("menu.bmp");
  gMenubg = SDL_ConvertSurface(temp, gScreen->format,SDL_SWSURFACE);
  SDL_FreeSurface(temp);

  temp = SDL_LoadBMP("fona.bmp");
  gFont = SDL_ConvertSurface(temp, gScreen->format,SDL_SWSURFACE);
  SDL_FreeSurface(temp);

  SDL_SetColorKey(gFont, SDL_SRCCOLORKEY, RGB(255,255,0));  

  genlevel(0, 0, 14);


  // If we fail, return error.
  if ( gScreen == NULL ) 
  {
    fprintf(stderr, "Unable to set 320x240 video: %s\n", SDL_GetError());
    exit(1);
  }

  // Main loop: loop forever.
  while (1)
  {
    // Render stuff
    render();

    // Poll for events, and handle the ones we care about.
    SDL_Event event;
    while (SDL_PollEvent(&event)) 
    {
      switch (event.type) 
      {
      case SDL_JOYBUTTONUP:
          handle_key(event.jbutton.button,0);
          break;
      case SDL_JOYBUTTONDOWN:
          handle_key(event.jbutton.button,1);
          break;
//        if (event.jbutton.button == GP2X_BUTTON_START ||
//            event.jbutton.button == GP2X_BUTTON_SELECT)
//          return 0;
      case SDL_KEYDOWN:
          handle_key(event.key.keysym.sym,1);
        break;
      case SDL_KEYUP:
          handle_key(event.key.keysym.sym,0);
        // If escape is pressed, return (and thus, quit)
//        if (event.key.keysym.sym == SDLK_ESCAPE)
//          return 0;
        break;
      case SDL_QUIT:
        return(0);
      }
    }
  }
  return 0;
}
