/*
Copyright (c) 2008, Jari Komppa

All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, 
  this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation 
  and/or other materials provided with the distribution.

* Neither the name of the library's copyright owner nor the names of its 
  contributors may be used to endorse or promote products derived from this 
  software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include "nelumbo.h"

struct menucontent
{
    int selected;
    int optcount;
    const char **text;
};

const char *menuquit[] = {"QUIT"};
const char *menurace[] = {
    "CUSTOM RACE",
    "DIFFICULT RACE",     
    "MEDIUM RACE", 
    "EASY RACE", 
    "QUICK RACE"    
};
const char *menushift[] = {"MANUAL SHIFT", "AUTO SHIFT"};
const char *menucartype[] = {"ACCELERATION","BALANCED","TOP SPEED"};
const char *menuaudio[] = {"MUSIC+SFX", "MUSIC ONLY", "SFX ONLY", "NO AUDIO"};
menucontent menu[] =
{
{0,1,menuquit},
{4,5,menurace},
{1,2,menushift},
{1,3,menucartype},
{0,4,menuaudio}
};
int menuitems = sizeof(menu) / sizeof(menucontent);

int menupick = 0;

void render_menu()
{   
  // Lock surface if needed
  if (SDL_MUSTLOCK(gScreen))
    if (SDL_LockSurface(gScreen) < 0) 
      return;

  int i, j;

  // Ask SDL for the time in milliseconds
  unsigned int tick = SDL_GetTicks(); 
  int menuofs = -95 + 170 * menupick;
  static FLOAT visualofs = 0;

  while (gLastTick < tick)
  {
      visualofs = (visualofs * 15 + menuofs) / 16;
      gLastTick += 10;
  }

	SDL_Rect src, dst;
	src.w = 320;
    src.h = 240;
	src.x = 0;
	src.y = 0;
	dst.w = 320;
	dst.h = 240;
	dst.x = 0;
	dst.y = 0;
	SDL_BlitSurface(gMenubg, &src, gScreen, &dst);        

    drawstring(VERSION,320-44,220);
    
    for (i = 0; i < menuitems; i++)
    {
        for (j = 0; j < menu[i].optcount; j++)
            drawstring(menu[i].text[j], i * 170 -TOINT(visualofs), 178 + (j - menu[i].selected) * 24);
    }
    
    drawstring("<               >",80,178);

  // Unlock if needed
  if (SDL_MUSTLOCK(gScreen)) 
    SDL_UnlockSurface(gScreen);

   // Tell SDL to update the whole gScreen
  SDL_UpdateRect(gScreen, 0, 0, 320, 240);    
}

void key_menu(int key, int down)
{
    if (down)
    {
        switch (key)
        {
        case GP2X_BUTTON_LEFT:
        case SDLK_LEFT:
            menupick = (menupick + menuitems - 1) % menuitems;
            break;
        case GP2X_BUTTON_RIGHT:
        case SDLK_RIGHT:
            menupick = (menupick + 1) % menuitems;
            break;
        case GP2X_BUTTON_UP:
        case SDLK_UP:
            menu[menupick].selected = (menu[menupick].selected + menu[menupick].optcount - 1) % menu[menupick].optcount;
            break;
        case GP2X_BUTTON_DOWN:
        case SDLK_DOWN:
            menu[menupick].selected = (menu[menupick].selected + 1) % menu[menupick].optcount;
            break;
        }
    }
    else
    {
        switch (key)
        {
        case GP2X_BUTTON_R:
        case GP2X_BUTTON_L:
        case GP2X_BUTTON_A:
        case GP2X_BUTTON_Y:
        case SDLK_RETURN:
        case SDLK_SPACE:
        case SDLK_LCTRL:
        case SDLK_RCTRL:
            if (menupick == 0)
                exit(0);
            if (menupick == 1)
                gGameState = 1;
            break;
        case GP2X_BUTTON_START:
        case GP2X_BUTTON_SELECT:
        case SDLK_ESCAPE:
            if (menupick == 0)
                exit(0);
            else
                menupick = 0;
            break;
        }

    }
}
