/*
Copyright (c) 2008, Jari Komppa

All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, 
  this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation 
  and/or other materials provided with the distribution.

* Neither the name of the library's copyright owner nor the names of its 
  contributors may be used to endorse or promote products derived from this 
  software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdlib.h>
#if defined(_MSC_VER)
#include "SDL.h"
#else
#include "SDL/SDL.h"
#endif
#include <math.h>
#include <string.h>
#ifndef _MSC_VER
#include <unistd.h>
#endif

#include "fixed_class.h"
#include "vector_math.h"
#include "mersennetwister.h"

using namespace fixedpoint;
using namespace vmath;

#define GP2X_BUTTON_UP              (0)
#define GP2X_BUTTON_DOWN            (4)
#define GP2X_BUTTON_LEFT            (2)
#define GP2X_BUTTON_RIGHT           (6)
#define GP2X_BUTTON_UPLEFT          (1)
#define GP2X_BUTTON_UPRIGHT         (7)
#define GP2X_BUTTON_DOWNLEFT        (3)
#define GP2X_BUTTON_DOWNRIGHT       (5)
#define GP2X_BUTTON_CLICK           (18)
#define GP2X_BUTTON_A               (12)
#define GP2X_BUTTON_B               (13)
#define GP2X_BUTTON_X               (14)
#define GP2X_BUTTON_Y               (15)
#define GP2X_BUTTON_L               (10)
#define GP2X_BUTTON_R               (11)
#define GP2X_BUTTON_START           (8)
#define GP2X_BUTTON_SELECT          (9)
#define GP2X_BUTTON_VOLUP           (16)
#define GP2X_BUTTON_VOLDOWN         (17)

#ifndef M_PI
#define M_PI 3.1415926535897932384626433832795f
#endif

#define VERSION "V0.2"

#define RGB(r,g,b) ((((r >> 3) & 0x1f) << 11) | (((g >> 2) & 0x3f) << 5) | ((b >> 3) & 0x1f))
#define COLORMIX(a,b,c) (((a)*(0xff-(c))+(b)*(c)) >> 8)

//#define CONSOLE_DEBUG

#ifdef CONSOLE_DEBUG
#define TIMING_START START = SDL_GetTicks();
#define TIMING_END(x) printf(x " %dms", SDL_GetTicks()-START);
#else
#define TIMING_START 
#define TIMING_END(x)
#endif

#define USE_FLOATS

#ifdef USE_FLOATS
typedef float FLOAT;
#define FLOOR(x) (floor(x))
#define TOINT(x) ((int)floor(x))
#define TOFLOAT(x) (x)
//#define sin(x) (sin(fixed_point<16>(x)).intValue * (1.0f / 65536.0f))
//#define cos(x) (cos(fixed_point<16>(x)).intValue * (1.0f / 65536.0f))
#else
typedef fixed_point<16> FLOAT;
#define FLOOR(x) FLOAT((x).intValue >> 16)
#define TOINT(x) ((x).intValue >> 16)
#define TOFLOAT(x) ((x).intValue * (1.0f / 65536.0f))
#endif

struct spritedata
{
    FLOAT x,y,d,w;
    SDL_Surface *pic;
};

struct leveldata
{
    int tracklen;
    int controlpoints;
    FLOAT *trackdata;
    FLOAT *trackseglen;    
    int staticsprites;
    spritedata *staticsprite;
    int mobilesprites;
    spritedata *mobilesprite;
    SDL_Surface *map;
    FLOAT mapscale;
    FLOAT mapxofs;
    FLOAT mapzofs;
    
    leveldata();
    ~leveldata();
};

extern leveldata *currentlevel;


extern int gGameState;

extern MersenneTwister gRand;
extern unsigned int gLastTick;
extern void drawstring(const char * string, int x, int y);

extern SDL_Surface *gScreen;
extern SDL_Surface *gBg;
extern SDL_Surface *gTreetest;
extern SDL_Surface *gCartest[3];
extern SDL_Surface *gFont;
extern SDL_Surface *gMenubg;
extern SDL_Surface *gCloudtest;
extern SDL_Surface *gGoaltest;

extern FLOAT bspline(FLOAT t, FLOAT p0, FLOAT p1, FLOAT p2, FLOAT p3);
extern FLOAT catmullrom(FLOAT t, FLOAT p0, FLOAT p1, FLOAT p2, FLOAT p3);
extern FLOAT linear(FLOAT t, FLOAT p0, FLOAT p1, FLOAT p2, FLOAT p3);
extern void trackpick(FLOAT &x, FLOAT &y, FLOAT &z, FLOAT d);
extern void trackpicklin(FLOAT &x, FLOAT &y, FLOAT &z, FLOAT d);

extern void render_menu();
extern void render_ingame();

extern void key_ingame(int key, int down);
extern void key_menu(int key, int down);

extern void genlevel(int basetype, int scenario, int seed);

